library ExtCommand;
{$apptype console}
uses
  Windows,
  FARIntf,
  ExtCommandMsg,
  Dirs, SysUtils, FARUtils, Classes, ConsoleApp, RegExpr;
type
  TInitDialogItem = packed record
    ItemType: integer;
    X1, Y1, X2, Y2: integer;
    Focus: integer;
    Selected: integer;
    Flags: DWORD;
    DefaultButton: integer;
    Data: PChar;
  end;
  TInitDialogItemArr = packed array [0..Pred(MaxInt div SizeOf(TInitDialogItem))] of TInitDialogItem;
  PInitDialogItemArr = ^TInitDialogItemArr;

const PluginCfg='ExtCommand.cfg'; // Config file name
      ExtExt='.ini'; // Extension for extension config
      AllFilesMenu='AllFiles'+ExtExt; //Menu name for all files
      FileNames:TStrings=Nil;
      title_changed:boolean=False;
var prev_title: PChar;

// Returns file from types dir
function typeFile(const s:String):String;
begin
  Result:=PluginDirFile('Types\'+s);
end;

procedure msg(Const s:string);
begin
  Message(FMSG_WARNING, 'Contents',
          [Integer(MError),
           s,
           Integer(MOk)
          ],
          1);
end;

// Returns right substring of S, starting with character number No
function rightStr(const s:string;No:Integer):string;
begin
  result:=copy(s, No, Length(s)-No+1);
end;

// Returns right substring of S, with length l
function right(const s:string;l:Integer):string;
begin
  result:=rightStr(s, Length(s)-l+1);
end;

procedure InitDialogItems(Init: PInitDialogItemArr; Item: PFarDialogItemsArray;
                          ItemsNumber: integer);
var
  i: integer;
begin
  for i:=0 to ItemsNumber-1 do begin
    integer(Item^[I].Typ):=Init^[I].ItemType;
    Item^[I].X1:=Init^[I].X1;
    Item^[I].Y1:=Init^[I].Y1;
    Item^[I].X2:=Init^[I].X2;
    Item^[I].Y2:=Init^[I].Y2;
    integer(Item^[I].Focus):=Init^[I].Focus;
    Item^[I].Selected:=Init^[I].Selected;
    Item^[I].Flags:=Init^[I].Flags;
    integer(Item^[I].DefaultButton):=Init^[I].DefaultButton;
    if (cardinal(Init^[I].Data)<2000) then
      StrCopy(Item^[I].Data, GetMsg(TMessageNumber(Init[I].Data)))
    else
      StrCopy(Item^[I].Data, Init^[I].Data);
  end;
end;

function SaveConsoleTitle: boolean;
begin
  GetMem(prev_title,512);
  Result:=GetConsoleTitle(prev_title,511)<>0;
end;

function RestoreConsoleTitle: boolean;
begin
  Result:=False;
  if not(prev_title=Nil) then begin
    Result:=SetConsoleTitle(prev_title);
    FreeMem(prev_title);
    prev_title:=Nil;
  end;
end;
{Register filename by EditorID}
procedure regFile(EditorID:integer;const FullPath:string);
Var ind:integer;
begin
  if FileNames=nil then
    FileNames:=TStringList.Create;
  ind:=FileNames.IndexOfObject(TObject(EditorID));
  if ind=-1 then
    FileNames.addObject(FullPath, TObject(EditorID))
  else
    FileNames[ind]:=FullPath;
end;

{Unregister filename by EditorID}
procedure unregFile(EditorID:integer);
Var ind:integer;
begin
  if FileNames<>nil then
  begin
    ind:=FileNames.IndexOfObject(TObject(EditorID));
    if ind<>-1 then
      FileNames.Delete(ind);
    if FileNames.Count=0 then
    begin
      FileNames.Free;
      FileNames:=nil;
    end;
  end;
end;

{Get registered file name by EditorID}
function getFileName(EditorID:integer):String;
Var ind:integer;
begin
  Result:='';
  if FileNames<>nil then
  begin
    ind:=FileNames.IndexOfObject(TObject(EditorID));
    if ind<>-1 then
      Result:= FileNames[ind];
  end;
end;
// whether S ends by S1
function ends(const S, S1:string):boolean;
begin
  if length(S)>=length(s1) then
  begin
    //msg(UpperCase(right(s, length(s1)))+'='+UpperCase(s1));
    Result:=UpperCase(right(s, length(s1)))=UpperCase(s1);
  end
  else
    Result:=false;
end;
function isSame(const FullPath, PartPath:string):boolean;
begin
  Result:=ends(FullPath, PartPath)
end;
function isLoaded(const FileName:String):boolean;
Var i:integer;
begin
  Result:=false;
  if FileNames<>nil then
  begin
    for i:=0 to FileNames.Count-1 do
    begin
      Result:=isSame(FileName, FileNames[i]);
      if Result then break;      
    end;  
  end;
end;

function currentlyEditing(const FileName:String):boolean;
var CurrentFullPath:string;
begin
  CurrentFullPath:=getFileName(EditorFileID);
  result:=isSame(CurrentFullPath, FileName)
end;
type TCommandOptions=class
     public 
       Save:boolean;
       ToMenu:boolean;
       RegExp:string;
       LineMatch:integer;
       FileMatch:integer;
       CDFileDir:boolean;
       AutoJump:boolean;
       AskTitle:string;
       constructor Create;
       function Clone:TCommandOptions;
     end;
constructor TCommandOptions.Create;
begin
  inherited;
  Save:=False;
  ToMenu:=False;
  RegExp:='';
  LineMatch:=0;
  FileMatch:=0;
  CDFileDir:=False;
  AutoJump:=False;
  AskTitle:='';
end;
function TCommandOptions.Clone;
begin
  Result:=TCommandOptions.Create;
  Result.Save:=Save;
  Result.ToMenu:=ToMenu;
  Result.RegExp:=RegExp;
  Result.LineMatch:=LineMatch;
  Result.FileMatch:=FileMatch;
  Result.CDFileDir:=CDFileDir;
  Result.AutoJump:=AutoJump;
  Result.AskTitle:=AskTitle;
end;

Var OutpOpt:TCommandOptions;
procedure EditorGo(LineNo:integer);overload;
begin
  SetEditorPosition(LineNo, 0, -1, LineNo, 0,  LongBool(-1));
end;
procedure EditorGo(const FileName, Title:string;LineNo:integer);overload;
begin
  if not (currentlyEditing(FileName) or (FileName=''))then
    Edit(FileName, Title,LineNo+1, 0)
  else
  begin
    EditorGo(LineNo);
    SaveConsoleTitle();
    SetConsoleTitle(PChar(Title));
  end;
end;

function ExtractRegExpNo(Var FileName:String;const Line:string):integer;
Var r:TRegExpr;
begin
  Result:=0;
  r:=TRegExpr.Create;
  try
    r.Expression:=OutpOpt.RegExp;
    if r.Exec(Line) then
    begin
      if r.SubExprMatchCount>=OutpOpt.LineMatch then
      begin
        Result:=StrToInt(r.Match[OutpOpt.LineMatch]);
      end;
      if r.SubExprMatchCount>=OutpOpt.FileMatch then
      begin
        FileName:=r.Match[OutpOpt.FileMatch];
      end;
    end;
  finally
    r.Free;
  end;
end;

function ExtractFirstNo(const Line:string):integer;
Var InNum:Boolean;
    i:Integer;
begin
  InNum:=false;
  Result:=0;
  for i:=1 to Length(Line) do
  begin
    if Line[i] in ['0'..'9'] then
    begin
      Result:=Result*10+Ord(Line[i])-Ord('0');
      InNum:=true;
    end;
    if InNum and Not (Line[i] in ['0'..'9']) then
      break;
  end;
end;

function ExtractNo(Var Filename:String;const Line:string):integer;
begin
  if OutpOpt.RegExp='' then
  begin
    Result:=ExtractFirstNo(Line);
    FileName:='';
  end
  else
    Result:=ExtractRegexpNo(FileName, Line);
end;

procedure EditorGo(const Line:string);overload;
Var lineNo:integer;
    FileName:String;
begin
  LineNo:=ExtractNo(FileName, Line);
  if LineNo>0 then
  begin
    EditorGo(FileName, Line, LineNo-1);
  end;
end;
// Parse single line onto appName and parameters
procedure parseCommandLine(const aLine:String;Var appName, params:string);
Var InQuote, InApp:Boolean;
    AppStart, AppEnd, i:integer;
begin
  InQuote:=false;
  InApp:=false;
  AppStart:=1;
  AppEnd:=Length(aLine)+1;
  for i:=1 to Length(aLine) do
  begin
    if (not inQuote) and (not InApp) then
    begin
      if aLine[i]='"' then
      begin
        inQuote:=true;
        AppStart:=i+1;
      end;
      if aLine[i]<>' ' then
      begin
        inApp:=True;
        AppStart:=i;
      end;
    end;
    if inQuote and (aLine[i]='"') then
    begin
      AppEnd:=i;
      break;
    end;    
    if inApp and (aLine[i]=' ') then
    begin
      AppEnd:=i;
      break;
    end;
  end;
  if AppStart<AppEnd then
  begin
    appName:=Copy(aLine, appStart, appEnd-appStart);
    params:=rightStr(aLine, appEnd+1);
  end
  else
  begin
    appName:=aLine;
    params:='';
  end;  
end;

procedure FreeStrings(a:TStrings);
Var i:Integer;
begin
  for i:=0 to a.Count-1 do
  begin
    if a.Objects[i]<>nil then
      a.Objects[i].Free;
  end;
  a.Free;
end;
type PFileOptions=^TFileOptions;
     TFileOptions=record
       Def:TCommandOptions;
     end;
procedure initFileOptions(o:PFileOptions);
begin
  o^.def:=TCommandOptions.Create;
end;
procedure destroyFileOptions(o:PFileOptions);
begin
  o^.Def.Free;
end;

function value(const S:String):String;
Var p:integer;
begin
  p:=Pos('=', S);
  result:=Copy(S, p+1, Length(S)-p);
end;
function isDirective(const S:String):boolean;
begin
  Result:=false;
  if s<>'' then Result:=S[1]='#';
end;
procedure setFileOptions(o:PFileOptions;S:TStrings);
Var i:integer;
    Cmd:String;
    CurOptions:TCommandOptions;
    function CurValue:String;
    begin
      Result:=value(S[i]);
    end;
    function CurUpValue:String;
    begin
      Result:=UpperCase(CurValue);
    end;
    function CurBool:boolean;
    begin
      Result:=CurUpValue='TRUE';
    end;
begin
  i:=0;
  CurOptions:=o^.Def;
  while i<S.Count do
  begin
    Cmd:=UpperCase(S.Names[i]);
    if isDirective(Cmd) then
    begin
      if Cmd='#SAVE' then
      begin
        CurOptions.Save:=CurBool;
      end;
      if Cmd='#TOMENU' then
      begin
        CurOptions.ToMenu:=CurBool;
      end;
      if Cmd='#REGEXP' then
      begin
        CurOptions.RegExp:=CurValue;
      end;
      if Cmd='#LINEMATCH' then
      begin
        CurOptions.LineMatch:=StrToInt(CurValue);
      end;
      if Cmd='#FILEMATCH' then
      begin
        CurOptions.FileMatch:=StrToInt(CurValue);
      end;
      if Cmd='#CDFILEDIR' then
      begin
        CurOptions.CDFileDir:=CurBool;
      end;
      if Cmd='#AUTOJUMP' then
      begin
        CurOptions.AutoJump:=CurBool;
      end;
      if Cmd='#ASK' then
      begin
        CurOptions.AskTitle:=CurValue;
      end;
      s.Delete(i);
    end
    else
    begin
      CurOptions:=o^.Def.Clone;
      s.Objects[i]:=CurOptions;
      inc(i);
    end;
  end;
end;
function ExecCmd(const Command:String;Menu:Boolean):TStrings;
Var c, p, cmd:String;
begin
  Result:=nil;
  cmd:=ExpandEnv(Command);
  if Menu then
  begin
    Result:=TStringList.Create;
    try
      parseCommandLine(cmd, c, p);
      SaveConsoleTitle();
      SetConsoleTitle(PChar(GetMsg(MPluginMenuString)+' - '+Command));
      ExecConsoleApp(c,P, Result,Nil);
      RestoreConsoleTitle();
    except
      on e:Exception do
      begin
        Result.Free;
        raise;
      end;
    end;
  end
  else
    FARExec(cmd);
end;
const Outp:TStrings=nil;
Var   OutpEd:integer;
      OutpCmd:String;
procedure SetOutp(op:TStrings;Ed:Integer;Const Cmd:String;Opt:TCommandOptions);
begin
  if Outp<>nil then
  begin
    Outp.Free;
    OutpOpt.Free;
  end;
  Outp:=op;
  OutpEd:=Ed;
  OutpOpt:=Opt;
  OutpCmd:=Cmd;
end;
procedure ShowOutp;
Var FileName:String;
    cnt,LineNo:integer;
begin
  LineNo:=0;
  if Outp<>nil then
  begin
    if OutpOpt.AutoJump then begin
      For cnt:=0 To Outp.Count-1 Do begin
        LineNo:=ExtractRegExpNo(FileName, Outp.Strings[cnt]);
        if LineNo>0 then begin
          EditorGo(FileName, Outp.Strings[cnt], LineNo-1);
          Break;
        end;
      end;
      if LineNo=0 then //no errors found
        EditorGo(SelectString(PChar(OutpCmd),Outp));
    end
    else begin //AutoJump=false
      EditorGo(SelectString(PChar(OutpCmd),Outp));
    end;
  end
  else //Outp=NIL
    Message(FMSG_WARNING, 'Contents',
          [Integer(MError),
           Integer(MEmptyOutput),
           Integer(MOk)
          ],
          1);
end;
procedure FreeOutp;
begin
  if Outp<>nil then
  begin
    Outp.Free;
    Outp:=nil;
    OutpOpt.Free;
  end;
  OutpCmd:='';
  OutpEd:=0;
end;

function AskIt(Title: string): string;
var
  Items: PFarDialogItemsArray;
  i: integer;
const
  itemsnum=6;
  initarray: packed array [0..itemsnum-1] of TInitDialogItem = (
{
           1         2         3         4
 01234567890123456789012345678901234567890123456789
0
1   |------------------------------------------|
2   | Ask title:                               |
3   | [______________________________________] |
4   |------------------------------------------|
5   |            [ OK ]   [ Cancel ]           |
6   |------------------------------------------|
7
}
  (ItemType:integer(DI_DOUBLEBOX); X1: 3; Y1: 1; X2:46; Y2: 6; Focus:0; Selected:0; Flags:0; DefaultButton:0; Data:PChar('')),
  (ItemType:integer(DI_TEXT);      X1: 5; Y1: 2; X2: 0; Y2: 0; Focus:0; Selected:0; Flags:0; DefaultButton:0; Data:PChar('')),
  (ItemType:integer(DI_EDIT);      X1: 5; Y1: 3; X2:44; Y2: 0; Focus:1; Selected:0; Flags:DIF_HISTORY; DefaultButton:0; Data:PChar('')),
  (ItemType:integer(DI_TEXT);      X1: 0; Y1: 4; X2: 0; Y2: 0; Focus:0; Selected:0; Flags:DIF_SEPARATOR; DefaultButton:0; Data:PChar('')),
  (ItemType:integer(DI_BUTTON);    X1: 0; Y1: 5; X2: 0; Y2: 0; Focus:0; Selected:0; Flags:DIF_CENTERGROUP; DefaultButton:1; Data:PChar(MOK)),
  (ItemType:integer(DI_BUTTON);    X1: 0; Y1: 5; X2: 0; Y2: 0; Focus:0; Selected:0; Flags:DIF_CENTERGROUP; DefaultButton:0; Data:PChar(MCancel))
  );
begin
  Result:='';
  if Length(Trim(Title))>0 then begin
    if Length(Title)>40 then SetLength(Title,40);
    initarray[2].Selected:=integer(PChar('extcom_ask'));
    initarray[1].Data:=PChar(Title);
    GetMem(Items,SizeOf(TFarDialogItem)*itemsnum);
    InitDialogItems(@initarray, Items, itemsnum);
    i:=Info.Dialog(Info.ModuleNumber,-1,-1,50,8,Nil,Items,itemsnum);
    if (i=4) then Result:=string(Items^[2].Data);
    FreeMem(Items);
  end;
end;

function OpenPlugin(OpenFrom: OPEN; Item: Integer): THandle stdcall export;
Var ei:TEditorInfo;
    menuFile, command:String;
    ss, all:TStrings;
    menuItems:PFARMenuItemsArray;
    choice:integer;
    fDir, fNameOnly,  fExt, fName, S:string;
    FullFileName:String;
    o:TFileOptions;
begin
  initFileOptions(@o);
  try
  all:=TStringList.Create;
  all.LoadFromFile(typeFile(AllFilesMenu));
  setFileOptions(@o, all);
  ss:=TStringList.Create;
  info.EditorControl(ECTL_GETINFO, @ei);
  MenuFile:=DirFile(getFileDir(ei.FileName), rightStr(EditorFileExt, 2)+ExtExt);
  if not FileExists(MenuFile) then
    MenuFile:=typeFile(rightStr(EditorFileExt, 2)+ExtExt);
  if FileExists(MenuFile) then
  begin
    ss.LoadFromFile(MenuFile);
    setFileOptions(@o, ss);
  end;
  S:=StrPas(GetMsg(MViewOutput));
  ss.insert(0, S+'=1');
  ss.insert(1, '---');
  ss.AddStrings(all);
  all.Free;
  if ss.Count>0 then
  begin
    menuItems:=newFARMenuItemsFromNames(ss);
    choice:=showFARMenu(GetMsg(MPluginMenuString), menuItems, ss.Count);
    freeFARMenuItems(menuItems, ss.Count);
    if Choice=0 then
      ShowOutp;
    if choice>=2 then
    begin
      if TCommandOptions(ss.Objects[choice]).Save then
        farSaveFile;
      info.EditorControl(ECTL_GETINFO, @ei);
      FullFileName:=getFileName(ei.EditorID);
      if FullFileName='' then
        FullFileName:=EditorFileFullPath;
      fDir:=ExtractFileDir(FullFileName);
      fName:=extractFileName(FullFileName);
      fNameOnly:=Copy(fName, 1, Length(fName)-Length(ExtractFileExt(fName)));
      fExt:=rightStr(ExtractFileExt(FullFileName), 2);
      command:=Format(ss.Values[ss.Names[choice]],
        [fullFileName, fDir, fName, fNameOnly, fExt, selection,
         currentWord, getWord, AskIt((TCommandOptions(ss.Objects[choice]).AskTitle))]);
      if TCommandOptions(ss.Objects[choice]).CDFileDir then
        ChDir(fDir);
      SetOutp(ExecCmd(command,
        TCommandOptions(ss.Objects[choice]).ToMenu),ei.EditorID,
        command, TCommandOptions(ss.Objects[choice]).Clone
      );
      if TCommandOptions(ss.Objects[choice]).ToMenu then
        ShowOutp;
    end;
  end;
  FreeStrings(ss);
  except
    on e:Exception do begin
        Message(FMSG_WARNING, 'Contents',
          [Integer(MError),
           e.Message,
           Integer(MOk)
          ],
          1);
    end;
  end;
  destroyFileOptions(@o);
  Result := INVALID_HANDLE_VALUE;
end;

function ProcessEditorInput(Rec: PInputRecord):integer; stdcall; export;
begin
  if (Rec.EventType=KEY_EVENT) and (Rec.Event.KeyEvent.bKeyDown) then
    RestoreConsoleTitle();
  Result:=0;
end;

function ProcessEditorEvent( Event:EEvent; Param:Pointer):integer;
  stdcall export;
Var ei:TEditorInfo;
    fullFileName:String;
begin
  // if closing ...
  if Event=EE_CLOSE then
  begin
    // unregistering file
    unregFile(PInteger(Param)^);
    if OutpEd=PInteger(Param)^ then
      FreeOutp;
  end;
  //if reading or saving
  if(Event = EE_READ) or (Event=EE_SAVE) then
  begin
    //registering file
    Info.EditorControl(ECTL_GETINFO,@ei);
    fullFileName:=getFullPath(ei.FileName);
    regFile(ei.EditorID, fullFileName);
  end;
  Result:=0;
end;

var PluginStrings: array[0..0] of PChar;

procedure GetPluginInfo(var Info: TPluginInfo);
  stdcall export;
begin
  with Info do
  begin
    StructSize := SizeOf(Info);
    Flags := PF_EDITOR or PF_DISABLEPANELS;

    DiskMenuStringsNumber := 0;

    PluginStrings[0]:=GetMsg(MPluginMenuString);
    PluginMenuStrings:=@PluginStrings;
    PluginMenuStringsNumber:=1;

    PluginConfigStringsNumber:=0;

    CommandPrefix:=nil;
  end;
end;
exports
  OpenPlugin,
  GetPluginInfo,
  ProcessEditorInput,
  ProcessEditorEvent;
begin
end.
